@@ -34,11 +34,7 @@ def list_employees(employee_generator, output_format, additional_header_items=No
             employee_list.append(employee)
     if employee_list:
         formatter = OutputFormatter(output_format, header)
-        if len(employee_list) > 10:
-            output = formatter.get_formatted_output(employee_list)
-            click.echo_via_pager(output)
-        else:
-            formatter.echo_formatted_list(employee_list)
+        formatter.echo_formatted_list(employee_list)
     else:
         click.echo("No users found.")
 
