@@ -3,6 +3,7 @@ import functools
 import click
 from pandas import DataFrame
 from pandas import json_normalize
+from py42.exceptions import Py42BadRequestError
 from py42.exceptions import Py42NotFoundError
 
 from code42cli.bulk import generate_template_cmd_factory
@@ -27,7 +28,10 @@ org_uid_option = click.option(
     help="Limit users to only those in the organization you specify. Note that child orgs are included.",
 )
 active_option = click.option(
-    "--active", is_flag=True, help="Limits results to only active users.", default=None,
+    "--active",
+    is_flag=True,
+    help="Limits results to only active users.",
+    default=None,
 )
 inactive_option = click.option(
     "--inactive",
@@ -215,6 +219,8 @@ _bulk_user_move_headers = ["username", "org_id"]
 
 _bulk_user_roles_headers = ["username", "role_name"]
 
+_bulk_user_alias_headers = ["username", "alias"]
+
 
 @users.command(name="move")
 @username_option("The username of the user to move.", required=True)
@@ -226,6 +232,39 @@ def change_organization(state, username, org_id):
     _change_organization(state.sdk, username, org_id)
 
 
+@users.command()
+@click.argument("username")
+@click.argument("alias")
+@sdk_options()
+def add_alias(state, username, alias):
+    """Add a cloud alias for a given user.
+
+    A cloud alias is the username an employee uses to access cloud services such as Google Drive or Box. Adding a cloud alias allows Incydr to link a user's cloud activity with their Code42 username. Each user has a default cloud alias of their Code42 username. You can add one additional alias."""
+    _add_cloud_alias(state.sdk, username, alias)
+
+
+@users.command()
+@click.argument("username")
+@click.argument("alias")
+@sdk_options()
+def remove_alias(state, username, alias):
+    """Remove a cloud alias for a given user."""
+    _remove_cloud_alias(state.sdk, username, alias)
+
+
+@users.command()
+@click.argument("username")
+@sdk_options()
+def list_aliases(state, username):
+    """List the cloud aliases for a given user.  Each user has a default cloud alias of their Code42 username with up to one additional alias."""
+    user = _get_user(state.sdk, username)
+    aliases = user["cloudUsernames"]
+    if aliases:
+        click.echo(aliases)
+    else:
+        click.echo(f"No cloud aliases for user '{username}' found.")
+
+
 @users.group(cls=OrderedGroup)
 @sdk_options(hidden=True)
 def orgs(state):
@@ -252,7 +291,8 @@ def _get_orgs_header():
 @format_option
 @sdk_options()
 def list_orgs(
-    state, format,
+    state,
+    format,
 ):
     """List all orgs."""
     pages = state.sdk.orgs.get_all()
@@ -269,7 +309,9 @@ def list_orgs(
 @format_option
 @sdk_options()
 def show_org(
-    state, org_uid, format,
+    state,
+    org_uid,
+    format,
 ):
     """Show org details."""
     formatter = OutputFormatter(format)
@@ -292,6 +334,8 @@ users_generate_template = generate_template_cmd_factory(
     commands_dict={
         "update": _bulk_user_update_headers,
         "move": _bulk_user_move_headers,
+        "add-alias": _bulk_user_alias_headers,
+        "remove-alias": _bulk_user_alias_headers,
     },
     help_message="Generate the CSV template needed for bulk user commands.",
 )
@@ -539,6 +583,86 @@ def bulk_remove_roles(state, csv_rows, format):
     formatter.echo_formatted_list(result_rows)
 
 
+@bulk.command(
+    name="add-alias",
+    help=f"Add aliases to a list of users from the provided CSV in format: {','.join(_bulk_user_alias_headers)}.\n\nA cloud alias is the username an employee uses to access cloud services such as Google Drive or Box. Adding a cloud alias allows Incydr to link a user's cloud activity with their Code42 username. Each user has a default cloud alias of their Code42 username. You can add one additional alias.",
+)
+@read_csv_arg(headers=_bulk_user_alias_headers)
+@format_option
+@sdk_options()
+def bulk_add_alias(state, csv_rows, format):
+    """Bulk add aliases to users"""
+
+    # Initialize the SDK before starting any bulk processes
+    # to prevent multiple instances and having to enter 2fa multiple times.
+    sdk = state.sdk
+    success_header = "alias added"
+
+    csv_rows[0][success_header] = "False"
+    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
+    stats = create_worker_stats(len(csv_rows))
+
+    def handle_row(**row):
+        try:
+            _add_cloud_alias(
+                sdk, **{key: row[key] for key in row.keys() if key != success_header}
+            )
+            row[success_header] = "True"
+        except Exception as err:
+            row[success_header] = f"False: {err}"
+            stats.increment_total_errors()
+        return row
+
+    result_rows = run_bulk_process(
+        handle_row,
+        csv_rows,
+        progress_label="Adding aliases to users:",
+        stats=stats,
+        raise_global_error=False,
+    )
+    formatter.echo_formatted_list(result_rows)
+
+
+@bulk.command(
+    name="remove-alias",
+    help=f"Remove aliases from a list of users from the provided CSV in format: {','.join(_bulk_user_alias_headers)}",
+)
+@read_csv_arg(headers=_bulk_user_alias_headers)
+@format_option
+@sdk_options()
+def bulk_remove_alias(state, csv_rows, format):
+    """Bulk remove aliases from users"""
+
+    # Initialize the SDK before starting any bulk processes
+    # to prevent multiple instances and having to enter 2fa multiple times.
+    sdk = state.sdk
+    success_header = "alias removed"
+
+    csv_rows[0][success_header] = "False"
+    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
+    stats = create_worker_stats(len(csv_rows))
+
+    def handle_row(**row):
+        try:
+            _remove_cloud_alias(
+                sdk, **{key: row[key] for key in row.keys() if key != success_header}
+            )
+            row[success_header] = "True"
+        except Exception as err:
+            row[success_header] = f"False: {err}"
+            stats.increment_total_errors()
+        return row
+
+    result_rows = run_bulk_process(
+        handle_row,
+        csv_rows,
+        progress_label="Removing aliases from users:",
+        stats=stats,
+        raise_global_error=False,
+    )
+    formatter.echo_formatted_list(result_rows)
+
+
 def _add_user_role(sdk, username, role_name):
     user_id = _get_legacy_user_id(sdk, username)
     _get_role_id(sdk, role_name)  # function provides role name validation
@@ -666,3 +790,21 @@ def _deactivate_user(sdk, username):
 def _reactivate_user(sdk, username):
     user_id = _get_legacy_user_id(sdk, username)
     sdk.users.reactivate(user_id)
+
+
+def _get_user(sdk, username):
+    # use when retrieving the user information from the detectionlists module
+    try:
+        return sdk.detectionlists.get_user(username).data
+    except Py42BadRequestError:
+        raise UserDoesNotExistError(username)
+
+
+def _add_cloud_alias(sdk, username, alias):
+    user = _get_user(sdk, username)
+    sdk.detectionlists.add_user_cloud_alias(user["userId"], alias)
+
+
+def _remove_cloud_alias(sdk, username, alias):
+    user = _get_user(sdk, username)
+    sdk.detectionlists.remove_user_cloud_alias(user["userId"], alias)
